﻿using System;
using System.Collections.Generic;
using System.Configuration;
using System.Data;
using System.IO;
using System.Linq;
using System.Text;
using System.Xml;
using System.Xml.Serialization;
using Npgsql;
using NpgsqlTypes;
using TapfloDB.Data;
using System.Reflection;

namespace TapfloDB
{
    public class Server
    {
        private static string host = "localhost";
        private static string port = "5432";
        private static string db = "tapflo";
        
        private static string cstr;
        private static StreamWriter _log = null;

        public static NpgsqlConnection Conn { get; private set; }
        public static Version DBVersion { get; private set; }
        public static Version AppVersion { get; private set; }
        
        static Server(){
            Conn = null;

            //AppVersion
            AppVersion = Assembly.GetExecutingAssembly().GetName().Version;

            //Initialize Log
            string log = ConfigurationManager.AppSettings["LogToFile"];
            if (log != null && log.ToLower() == "true"){
                try {
                    log = ConfigurationManager.AppSettings["LogAppend"];
                    _log = new StreamWriter("TapfloDB.log", log != null && log.ToLower() == "true");
                    _log.AutoFlush = true;
                    Log("Uruchomiono TapfloDB {0} ({1})", AppVersion, GetShortOS());
                }
                catch {
                    _log = null;
                }
            }
        }

        ~Server(){
            if (_log != null)
                _log.Dispose();
        }

        public static bool CheckConfiguration(){
            //cstr = "Server=127.0.0.1;Port=5432;User Id=tapflodb;Password=tapflodb;Database=tapflo;"
            try {
                BuildDefaultConnStr();
            }
            catch {
                Log("Nie odnaleziono konfiguracji połączenia z serverem!");
                return false;
            }

            return true;
        }

        public static void Log(String format, params Object[] args){
            if (_log == null) return;

            string prefix = DateTime.Now.ToString("[dd.MM.yyyy HH:mm:ss] ");
            _log.WriteLine(prefix + format, args);
        }

        #region DB Connection

        private static void BuildConnStr(string login, string password){
            if(ConfigurationManager.AppSettings["Host"] != null)
                host = ConfigurationManager.AppSettings["Host"];
            if(ConfigurationManager.AppSettings["Port"] != null)
                port = ConfigurationManager.AppSettings["Port"];
            if(ConfigurationManager.AppSettings["DB"] != null)
                db = ConfigurationManager.AppSettings["DB"];
            
            if (ConfigurationManager.ConnectionStrings["TapfloDB"] != null)
                cstr = String.Format("{0}User Id={1};Password={2};",
                       ConfigurationManager.ConnectionStrings["TapfloDB"], login, password
                   );
            else
                cstr = String.Format("Server={0};Port={1};Database={2};User Id={3};Password={4};",
                       host, port, db, login, password
                   );
        }

        private static void BuildDefaultConnStr() {
            BuildConnStr(ConfigurationManager.AppSettings["DBUser"],
                         ConfigurationManager.AppSettings["DBPassword"]);
        }

        public static bool CheckConnection() {
            try {
                Connect();
                try {
                    DBVersion = new Version((string)Command("SELECT tapflodb();").ExecuteScalar());
                }
                catch { }
                Log("Połączono z {2} @ {0}:{1} (TapfloDB {3} @ Postgres {4})", host, port, db, DBVersion, Conn.ServerVersion);
            }
            catch (NpgsqlException exp){
                Log("Błąd podczas łączenia z {2} @ {0}:{1} - {3}", host, port, db, exp.Message);
                return false;
            }
            catch (Exception exp){
                Log("Nieznany błąd nawiązywania połączenia z serverem - {0}", exp.Message);
                return false;
            }

            return true;
        }

        public static NpgsqlConnection Connect(){
            if (Conn != null){
                if (Conn.State == ConnectionState.Open && Conn.FullState != ConnectionState.Broken)
                    return Conn;
                else {
                    Conn.Dispose();
                    Conn = null;
                }
            }
            Conn = new NpgsqlConnection(cstr);
            Conn.Open();

            return Conn;
        }

        public static void Close(){
            if (Conn != null)
                Conn.Close();
        }

        public static NpgsqlCommand Command(string query) {
            Connect();
            
            return new NpgsqlCommand(query, Conn);
        }

        public static NpgsqlDataAdapter DataAdapter(string query){
            Connect();

            return new NpgsqlDataAdapter(query, Conn);
        }

        public static string SetClause(params string[] set) {
            StringBuilder sb = new StringBuilder(set.Length * 10);
            foreach (string s in set)
                sb.AppendFormat("{1}{0} = :{0}", s, sb.Length > 0 ? ", " : "");
            
            return sb.ToString();
        }

        #endregion

        #region Categories/Products

        private List<Category> categs = new List<Category>();
        XmlSerializer categxs = new XmlSerializer(typeof(List<User>), new XmlRootAttribute("Categories"));
        public List<Category> GetCategories(){
            if (categs.Count > 0)
                return categs;

            TextReader tr = null;
            try {
                tr = new StreamReader(@"products.xml");
                categs = (List<Category>)categxs.Deserialize(tr);
            }
            finally {
                if(tr != null)
                    tr.Close();
            }

            return categs;
        }

        public void Add(Category c) {
            categs.Add(c);
        }

        public void SaveCategories(){
            TextWriter tw = new StreamWriter(@"products.xml");
            if(categs.Count > 0)
                categxs.Serialize(tw, categs);
            
            tw.Close();
        }

        #endregion

        #region User Login

        public static bool Login(string login, string pass){
            Connect();
            
            //Build Query
            NpgsqlCommand command = new NpgsqlCommand("select * from login(:login, :pass, :os, :account)", Conn);
            //Set params
            command.Parameters.AddWithValue("login", login);
            command.Parameters.AddWithValue("pass", pass.SHA256());
            command.Parameters.AddWithValue("os", GetShortOS() + " TapfloDB " + AppVersion);
            command.Parameters.AddWithValue("account", Environment.UserName);
            
            //Exec
            NpgsqlDataAdapter da = null;
            try {
                da = new NpgsqlDataAdapter(command);
                DataTable dt = new DataTable();
                da.Fill(dt);
                if (dt.Rows.Count == 0)
                    return false;

                Close(); //Close Connection
                BuildConnStr(login, pass);
                Connect(); //Conect as final user

                Session.Start(new User(dt.Rows[0]));
            }
            catch (NpgsqlException exp) {
                BuildDefaultConnStr();
                throw exp;
            }
            finally {
                if(da != null)
                    da.Dispose();
            }

            return true;
        }

        private static string GetShortOS() {
            Microsoft.VisualBasic.Devices.ComputerInfo ci = new Microsoft.VisualBasic.Devices.ComputerInfo();
            string name = ci.OSFullName.Replace("Microsoft ", "");
            OperatingSystem os = Environment.OSVersion;
            string sp = os.ServicePack.Replace("Dodatek", "").Replace("Service Pack ", "SP");
            string net = Assembly.GetExecutingAssembly().GetReferencedAssemblies().Where(x => x.Name == "System.Core").First().Version.ToString();
            return String.Format("{0}{1} ({2} {3}) .Net {4}", name, sp, os.Platform, os.Version, net);
        }

        public static bool Passwd(int uid, string oldpass, string newpass){
            Connect();

            //Build Query
            NpgsqlCommand command = new NpgsqlCommand("select changepassword(:uid, :oldpass, :newpass)", Conn);
            //Set params
            command.Parameters.AddWithValue("uid", uid);
            command.Parameters.AddWithValue("oldpass", oldpass.SHA256());
            command.Parameters.AddWithValue("newpass", newpass);

            return (int)command.ExecuteScalar() == 0;
        }

        #endregion

        #region User

        public static bool Toogle(int uid) {
            NpgsqlCommand cmd = Server.Command("UPDATE tapflodb.userdata SET mode = mode # 1 WHERE id = :id;");
            cmd.Parameters.AddWithValue("id", uid);

            return cmd.ExecuteNonQuery() > 0;
        }

        public static bool Delete(int uid, string login){
            NpgsqlConnection conn = Server.Connect();
            NpgsqlTransaction trans = conn.BeginTransaction();
            
            NpgsqlCommand cmd = Server.Command("DELETE FROM tapflodb.userdata WHERE id = :id;");
            cmd.Parameters.AddWithValue("id", uid);

            //Role
            NpgsqlCommand role = Server.Command(String.Format("DROP USER {0};", login));

            bool result = false;
            try {
                //Catch 42704
                try {
                    role.ExecuteNonQuery();
                }
                catch(NpgsqlException exp){
                    if (exp.Code != "42704") //Role not exist
                        throw exp;
                    else {
                        trans.Rollback();
                        trans = conn.BeginTransaction();
                    }
                }

                if (cmd.ExecuteNonQuery() == 1){
                    trans.Commit();
                    result = true;
                }
                else
                    trans.Rollback();
            }
            catch (Exception exp){
                trans.Rollback();
                Log("Błąd usuwania użytkownika {0} (#{1}) - {2}", login, uid, exp.Message);
            }

            return result;
        }

        #endregion
    }
}
