﻿using System;
using System.Configuration;
using System.Collections.Generic;
using System.Collections.Specialized;
using System.Drawing;
using System.Drawing.Printing;
using System.Text;
using System.Text.RegularExpressions;
using System.ComponentModel;

namespace TapfloDB.View.TDB
{
    class LabelPrinter : System.Drawing.Printing.PrintDocument, INotifyPropertyChanged
    {
        #region INotifyPropertyChanged

        public event PropertyChangedEventHandler PropertyChanged;
        protected virtual void OnPropertyChanged(string propertyName){
            PropertyChangedEventHandler handler = PropertyChanged;
            if (handler != null) handler(this, new PropertyChangedEventArgs(propertyName));
        }

        protected bool SetField<T>(ref T field, T value, string propertyName){
            if (EqualityComparer<T>.Default.Equals(field, value)) return false;
            field = value;
            OnPropertyChanged(propertyName);
            return true;
        }

        private static LabelPrinter instance = null;
        protected static bool SetStaticField<T>(ref T field, T value, string propertyName){
            if (EqualityComparer<T>.Default.Equals(field, value)) return false;
            field = value;
            if (instance != null)
                instance.OnPropertyChanged(propertyName);
            return true;
        }

        #endregion

        #region Fields

        // settings handle
        private static Properties.Settings settings = Properties.Settings.Default;

        // text layout of document
        private static StringFormat Format;

        // posible label headers
        private static string AODPHeader = "Air oparated diaphragm pump";
        private static Dictionary<string, string> Headers = new Dictionary<string, string>(){
            {"^CTI", "Centrifugal pump"},
            {"^CTH", "Centrifugal pump"},
            {"^CTM", "Magnetic driven centrifugal pump"},
            {"^CTV", "Vertical centrifugal pump"},
            {"^CTS", "Self-priming centrifugal pump"},
            {"^T",   "Air operated diaphragm pump"},
            {"^D",   "Pulsation dampener"},
            {"^FTA", "Filter Units"},
            {"^NM",  "Pneumixer"},
            {"^P",   "Pulsation dampener"}
        };

        // label header mode
        private static string[] modes = { "Auto", "AODP", "None" };
        private static string _mode = modes[0];

        #endregion

        #region Properties

        public static Font HeaderFont { get; set; }
        public static Font ModelFont { get; set; }
        public static Font SNFont { get; set; }

        public static RectangleF HeaderRect { get; set; }
        public static RectangleF ModelRect { get; set; }
        public static RectangleF SNRect { get; set; }

        public string Header { get; set; }
        public string Model { get; set; }
        public string SN { get; set; }

        // dummy property for binding
        public string HeaderMode
        {
            get { return _mode.Replace("None", "Brak"); }
        }

        public static string Mode 
        {
            get { return _mode; }
            set {
                string mode = value.Replace("Brak", "None");
                foreach (string m in modes){
                    if (String.Compare(m, mode, true) == 0){
                        SetStaticField(ref _mode, m, "HeaderMode");
                        settings.HeaderlMode = m;
                        settings.Save();
                    }
                }
            }
        }

        #endregion

        #region Class Constructors 

        static LabelPrinter() {
            Format = new StringFormat(StringFormatFlags.LineLimit);
            Format.Alignment = StringAlignment.Near; // left
            Format.LineAlignment = StringAlignment.Far; // bottom

            HeaderFont = new Font("Arial", 13f, FontStyle.Bold);
            ModelFont = new Font("Arial", 18f, FontStyle.Bold);
            SNFont = new Font("Arial", 14f, FontStyle.Bold);

            // First Try
            //ModelRect = new RectangleF(88.38f, 18.12f, 72f, 8f);
            //SNRect = new RectangleF(107.43f, 29.43f, 52f, 7f);
            // Adjusted
            //ModelRect = new RectangleF(79.38f, 12.12f, 72f, 8f);
            //SNRect = new RectangleF(98.43f, 23.43f, 52f, 7f);
            // 2014 Update
            // ModelRect = new RectangleF(79.5f, 12f, 72f, 8f);
            // SNRect = new RectangleF(98.5f, 24.5f, 52f, 7f);
            // 1.7 RC5
            LoadSettings();
        }

        /// <summary>
        /// Empty constructor
        /// </summary>
        public LabelPrinter() : base(){
            if (instance == null)
                instance = this;
        }

        /// <summary>
        /// Constructor to initialize our printing object
        /// and the text it's supposed to be printing
        /// </summary>
        public LabelPrinter(string model, string sn) : this(){
            Header = HeaderFromModel(model);
            Model = model;
            SN = sn;
        }

        public static void LoadSettings(){
            Mode = settings.HeaderlMode;
            HeaderRect = new RectangleF(settings.HeaderPosition, new SizeF(88f, 6f));
            ModelRect = new RectangleF(settings.ModelPosition, new SizeF(72f, 8f));
            SNRect = new RectangleF(settings.SNPosition, new SizeF(53f, 7f));
        }

        public static void LoadLegacySettings() {
            NameValueCollection s = ConfigurationManager.AppSettings;

            // 1.7 RC5 - 1.7 RC6
            Mode = s["LabelHeaderMode"];

            /* 1.7 RC4 -> 1.7 RC5
            if (LoadSetting("Model", s["ModelX"], s["ModelY"]))
                ModelRect = new RectangleF(settings.ModelPosition, ModelRect.Size);
            if (LoadSetting("SN", s["SNX"], s["SNY"]))
                SNRect = new RectangleF(settings.SNPosition, SNRect.Size);
            */
        }

        public static bool LoadSetting(string name, string arg1, string arg2) {
            bool update = false;
            string k = name + "Position";
            if (settings[k] == null)
                return false;

            // get previous/default values
            float x = ((PointF)settings[k]).X, y = ((PointF)settings[k]).Y;

            // try get new values
            if (arg1 != null && float.TryParse(arg1, out x))
                update = true;
            if (arg2 != null && float.TryParse(arg2, out y))
                update = true;

            if (update)
                settings[k] = new PointF(x, y);
            
            return update;
        }

        private static string HeaderFromModel(string model) { 
            if (Mode == "AODP")
                return AODPHeader;
            else if (Mode == "None")
                return "";

            foreach (KeyValuePair<string, string> head in Headers){
                if (Regex.Match(model, head.Key, RegexOptions.IgnoreCase).Success)
                    return head.Value;
            }

            return "";
        }

        #endregion

        #region PrintDocument Implementation

        /// <summary>
        /// Override the default OnBeginPrint method of the PrintDocument
        /// </summary>
        protected override void OnBeginPrint(PrintEventArgs e)
        {
            base.OnBeginPrint(e);

            DocumentName = "Etykieta " + Model;
        }

        /// <summary>
        /// Override the default OnPrintPage method of the PrintDocument
        /// </summary>
        protected override void OnPrintPage(System.Drawing.Printing.PrintPageEventArgs e){
            // Run base code
            base.OnPrintPage(e);

            //Declare local variables needed
            /*
            int printHeight;
            int printWidth;
            int leftMargin;
            int topMargin;

            printHeight = base.DefaultPageSettings.PaperSize.Height - base.DefaultPageSettings.Margins.Bottom;
            printWidth = base.DefaultPageSettings.PaperSize.Width - base.DefaultPageSettings.Margins.Right;
                
            leftMargin = base.DefaultPageSettings.Margins.Left;
            topMargin = base.DefaultPageSettings.Margins.Top;

            //Check if the user selected to print in Landscape mode
            //if they did then we need to swap height/width parameters
            if (base.DefaultPageSettings.Landscape){
                int tmp;
                tmp = printHeight;
                printHeight = printWidth;
                printWidth = tmp;
            }
            */

            PrintLabelElements(e.Graphics);

            e.HasMorePages = false;
        }

        #endregion

        #region Custom Print Functions

        private void PrintLabelElements(Graphics g){
            // use milimeters
            g.PageUnit = GraphicsUnit.Millimeter;

            // DebugBackground(g);

            // print label elements
            PrintAdjusted(g, Header, HeaderFont, Brushes.Black, HeaderRect);
            PrintAdjusted(g, Model, ModelFont, Brushes.Black, ModelRect);
            PrintAdjusted(g, SN, SNFont, Brushes.Black, SNRect);
            
            /*
            String num = "1234567890123456789012345678901234567890";
            for (int i = 0; i < 8; ++i){
                Brush b = new SolidBrush(Color.FromArgb(i*12, i*12, i*12));
                PrintAdjusted(g, Model + num.Substring(0, (int)(i * 2.5 + i)), ModelFont, b, ModelRect);
            }
            */
        }

        private void PrintAdjusted(Graphics g, string text, Font font, Brush brush, RectangleF rect) {
            bool fits = false;
            double limit = font.Size / 2f;
            Font f = font;
            SizeF size = SizeF.Empty;

            // find best font size
            while (!fits && f.Size > limit){
                // measure string to print
                size = g.MeasureString(text, f, 1000, Format);
                // Console.WriteLine("f: {0}; size: {1} x {2}", f.Size, size.Width, size.Height);
                
                if (size.Width < rect.Width)
                    fits = true;
                else
                    f = new Font(font.FontFamily, f.SizeInPoints - 2f, font.Style);
            }

            // calc new rect
            FontFamily ff = f.FontFamily;
            int fdst = ff.GetCellDescent(f.Style), div = ff.GetEmHeight(f.Style);
            double dp = font.SizeInPoints - f.SizeInPoints;
            double dy = (font.GetHeight(g) - f.GetHeight(g)) * fdst / div + (dp * 0.025);
            double dx = dp * 0.0625;
            RectangleF r = new RectangleF((float)(rect.X + dx), (float)(rect.Y - dy), rect.Width, rect.Height);

            // Draw string
            g.DrawString(text, f, brush, r, Format);
        }

        #endregion

        #region Debug Print Functions

        public void DebugPrint(string filename){
            // create image
            Bitmap bmp = new Bitmap(10000, 2000);
            Graphics g = Graphics.FromImage(bmp);
            g.PageScale = 10f;

            DebugBackground(g);
            PrintLabelElements(g);

            // save image
            bmp.Save(filename);
        }

        private void DebugBackground(Graphics g) {
            // black background
            g.FillRectangle(Brushes.Black, 0f, 0f, 200f, 50f);

            // white field placeholders
            g.FillRectangle(Brushes.Wheat, HeaderRect);
            g.FillRectangle(Brushes.Wheat, ModelRect);
            g.FillRectangle(Brushes.Wheat, SNRect);

            Font f = ModelFont;
            FontFamily ff = f.FontFamily;
            Console.WriteLine("eh: {0}; ad: {1}; cd: {2}; ls: {3}", ff.GetEmHeight(f.Style), ff.GetCellAscent(f.Style), ff.GetCellDescent(f.Style), ff.GetLineSpacing(f.Style));

            // draw grid (x)
            Pen n = new Pen(Color.DarkGreen, 0.02f);
            for (float y = ModelRect.Bottom - 1, x = ModelRect.Left + 0.5f; y > ModelRect.Y + 5f; y -= 0.05f, x += 0.1f){ 
                g.DrawLine(n, x, y, ModelRect.Right - 60f, y);
                g.DrawLine(n, x, y, x, ModelRect.Top + 1f);
            }

            // print font lines
            // DrawFontLines(g, Color.DarkGray, HeaderFont, HeaderRect);
            // DrawFontLines(g, Color.DarkGray, new Font(HeaderFont.FontFamily, HeaderFont.SizeInPoints - 2f, HeaderFont.Style), HeaderRect);

            // DrawFontLines(g, Color.DarkGray, ModelFont, ModelRect);
            // DrawFontLines(g, Color.DarkGray, new Font(ModelFont.FontFamily, ModelFont.SizeInPoints - 2f, ModelFont.Style), ModelRect);
            // DrawFontLines(g, Color.DarkGray, new Font(ModelFont.FontFamily, ModelFont.SizeInPoints - 6f, ModelFont.Style), ModelRect);
            
            // DrawFontLines(g, Color.DarkGray, SNFont, SNRect);
            // DrawFontLines(g, Color.DarkGray, new Font(SNFont.FontFamily, SNFont.SizeInPoints - 2f, SNFont.Style), SNRect);
        }

        #endregion
    }
}
